program rd0xab_delphi;

uses
  Windows, SysUtils;

{$APPTYPE CONSOLE}
{$WARNINGS OFF}
const
// ******************************************************
// *** public GLOBAL constant & vars for rd0xAB32.dll ***
// ******************************************************
    UDEV_REV            = $20;  // HW min.version for CCR

    SZ_UID              = 10;   // UID size for data frames
    SZ_SCCD             = 14;   // card descriptor size for data frames: uidlen(1)+UID(10)+SAK(1)+ATQA(2)

    SZ_MFBLK            = 16;   // block size in bytes
    SZ_MFPAGE           = 4;    // page size in bytes

    MFDM_BLOCK          = $00;  // block access to data (16b) if b.0 = 0
    MFDM_PAGE           = $01;  // page access to data (4b) if b.0 = 1
    MFDM_DM             = $00;  // data read mode if b.1 = 0
    MFDM_AM             = $02;  // authorization only mode (w/out read) if b.1 = 1
    MFDM_NMC            = $00;  // normal card operation if b.7 == 0
    MFDM_RWC            = $80;  // RW card operation if b.7 == 1

    SZ_ATS              = 30;   // size of ATS

// * name import from library *
    RD0XAB_NM           = 'rd0xAB32.dll';
    LINK_ISOPEN_NM      = 'link_isopen';
    LINK_CLOSE_NM       = 'link_close';
    LINK_HIDOPEN_NM     = 'link_hidopen';
    LINK_COMOPEN_NM     = 'link_comopen';
    LINK_FLUSH_NM       = 'link_flush';
    LINK_ECHOBYTE_NM    = 'link_echobyte';
    LINK_RQWAIT_NM      = 'link_rqwait';
    LINK_RQACK_NM       = 'link_rqack';
    LINK_PACKET_NM      = 'link_packet';
    ACQ2TYPE_NM         = 'acq2type';
    MFINIT_NM           = 'mfInit';
    MFBEEP_NM           = 'mfBeep';
    MFBRKCARRIER_NM     = 'mfBrkCarrier';
    MFTYPE_NM           = 'mfType';
    MFSELECT_NM         = 'mfSelect';
    MFSETAK_NM          = 'mfSetAK';
    MFAUTH_NM           = 'mfAuth';
    MFLONG_NM           = 'mfLong';
    MFCALCCRCA_NM       = 'mfCalcCrcA';
    MFAPPENDCRCA_NM     = 'mfAppendCrcA';
    MFIO_NM             = 'mfIO';
    MFBITIO_NM          = 'mfBitIO';
    MFDATAREAD_NM       = 'mfDataRead';
    MFDATAWRITE_NM      = 'mfDataWrite';
// * card type *
    SCD_UNKNOWN         = 0;    // ?
    SCD_UL              = $01;  // ultralight
    SCD_CL1K            = $02;  // Classic 1K
    SCD_CL4K            = $03;  // Classic 4K
    SCD_CL1K1           = $04;  // Classic 1K, Infineon
    SCD_CL4K1           = $05;  // Classic 4K, Infineon
    SCD_PL2S1           = $06;  // Plus 2K, Sec.Lev.1
    SCD_PL4S1           = $07;  // Plus 4K, Sec.Lev.1
    SCD_PL2S2           = $08;  // Plus 2K, Sec.Lev.2
    SCD_PL4S2           = $09;  // Plus 4K, Sec.Lev.2
    SCD_PL2S3           = $0A;  // Plus 2K, Sec.Lev.3
    SCD_PL4S3           = $0B;  // Plus 4K, Sec.Lev.3
    SCD_MINI            = $0C;  // Classic Mini
    SCD_DF1             = $0D;  // DesFire, EV1
    SCD_DF2             = $0E;  // DesFire, EV2
    SCD_SMMX            = $0F;  // Smart MX
    SCD_SM1K            = $10;  // Smart MX 1K
    SCD_SM2K            = $11;  // Smart MX 2K
    SCD_SM4K            = $12;  // Smart MX 4K
// * fn error codes *
    LERR_SUCCESS        = 0;
    LERR_FAIL           = 1;
    LERR_HARDWARE       = 2;
    LERR_TIMEOUT_TX     = 3;
    LERR_TIMEOUT_RX     = 4;
    LERR_NOTSYNC        = 5;
    LERR_NOREADER       = 6;    // card error codes (high level errors)
    LERR_NOCARD         = 7;
    LERR_INVALIDCARD    = 8;
    LERR_INVALIDKEY     = 9;
    LERR_IOERR          = 10;
    LERR_DATAREAD       = 11;
    LERR_DATAWRITE      = 12;
    LERR_BUSY           = 13;
    LERR_USER           = 14;
// * mifare commands, low level commands *
    PICC_REQIDL         = $26;  // request idle
    PICC_REQRW          = $40;  // request RW (mifare zero)
    PICC_REQALL         = $52;  // request all
    PICC_ANTICOLLRW     = $43;  // anticollision for RW-card (mifare zero)
    PICC_ANTICOLL1      = $93;  // anticollision level 1 106 kBaud
    PICC_ANTICOLL11     = $92;  // anticollision level 1 212 kBaud
    PICC_ANTICOLL12     = $94;  // anticollision level 1 424 kBaud
    PICC_ANTICOLL13     = $98;  // anticollision level 1 848 kBaud
    PICC_ANTICOLL2      = $95;  // anticollision level 2
    PICC_ANTICOLL3      = $97;  // anticollision level 3
    PICC_HALT           = $50;  // halt
    PICC_PWDAUTH        = $1B;  // password authentication
    PICC_GETVERSION     = $60;  // information on the MIFARE family
    PICC_AUTHENT1A      = $60;  // authentication using key A
    PICC_AUTHENT1B      = $61;  // authentication using key B
    PICC_READ16         = $30;  // read 16 byte block
    PICC_READCNT        = $39;  // read out counters value
    PICC_FASTREAD       = $3A;  // fast read
    PICC_READSIG        = $3C;  // read manufacturer signature
    PICC_CHKTEARING     = $3E;  // enables the application to identify
    PICC_VCSL           = $4B;  // enable a unique identification and selection
    PICC_WRITE16        = $A0;  // write 16 byte block
    PICC_WRITE4         = $A2;  // write 4 byte block
    PICC_INCRCNT        = $A5;  // increment one-way counters
    PICC_TRANSFER       = $B0;  // transfer command code
    PICC_DECREMENT      = $C0;  // decrement value
    PICC_INCREMENT      = $C1;  // increment value
    PICC_RESTORE        = $C2;  // restore command code
    PICC_RATS           = $E0;  // command identifies the PICC type

type
// * device pseudo-handle *
    TCCR = packed record        // reader's pseudo-handle struct
        hL:         THandle;    // handle
        txq:        DWORD;      // tx quantile
        rxout:      DWORD;      // rx timeout
        txout:      DWORD;      // tx timeout
        wtout:      DWORD;      // wait timeout
        inbuf:      PBYTE;      // in-queue
        inlen:      BYTE;       // length of in-queue
        outbuf:     PBYTE;      // out-queue
        outlen:     BYTE;       // length of out-queue
    end;
    PCCR = ^TCCR;
// * fn's samples *
    _s_link_isopen      = function(pccr:PCCR):DWORD; stdcall;
    _s_link_close       = procedure(pccr:PCCR); stdcall;
    _s_link_hidopen     = function(idx:DWORD; pusn:PDWORD; pccr:PCCR):DWORD; stdcall;
    _s_link_comopen     = function(port:DWORD; cbr:DWORD; pccr:PCCR):DWORD; stdcall;
    _s_link_flush       = function(pccr:PCCR):DWORD; stdcall;
    _s_link_echobyte    = function(pccr:PCCR; dt:BYTE):DWORD; stdcall;
    _s_link_rqwait      = function(pccr:PCCR; buf:PBYTE; len:BYTE):DWORD; stdcall;
    _s_link_rqack       = function(pccr:PCCR; ack:BYTE):DWORD; stdcall;
    _s_link_packet      = function(pccr:PCCR; cmd:BYTE; parm:BYTE; outbuf:PBYTE; outlen:WORD; inbuf:PBYTE; inlen:WORD):DWORD; stdcall;
    _s_acq2type         = function(acq:DWORD; atslen:BYTE; ats:PBYTE):BYTE; stdcall;
    _s_mfInit           = function(pccr:PCCR; pmodel:PBYTE; phw:PBYTE; psn:PDWORD):DWORD; stdcall;
    _s_mfBeep           = function(pccr:PCCR; tone:BYTE):DWORD; stdcall;
    _s_mfBrkCarrier     = function(pccr:PCCR; time_ms:WORD):DWORD; stdcall;
    _s_mfType           = function(pccr:PCCR; timeout:BYTE; ptype:PBYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE; patslen:PBYTE; pats:PBYTE):DWORD; stdcall;
    _s_mfSelect         = function(pccr:PCCR; timeout:BYTE; rq:BYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE):DWORD; stdcall;
    _s_mfSetAK          = function(pccr:PCCR; ak:BYTE; key:PBYTE):DWORD; stdcall;
    _s_mfAuth           = function(pccr:PCCR; timeout:BYTE; acm:BYTE; ak:BYTE; block:BYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE):DWORD; stdcall;
    _s_mfLong           = function(pccr:PCCR; txbuf:PBYTE; txlen:BYTE; ans:PBYTE; pabitlen:PBYTE):DWORD; stdcall;
    _s_mfCalcCrcA       = function(buf:PBYTE; len:DWORD):WORD; stdcall;
    _s_mfAppendCrcA     = procedure(buf:PBYTE; len:DWORD); stdcall;
    _s_mfIO             = function(pccr:PCCR; buf:PBYTE; len:BYTE; par:PBYTE; ans:PBYTE; pabitlen:PBYTE; anspar:PBYTE):DWORD; stdcall;
    _s_mfBitIO          = function(pccr:PCCR; buf:PBYTE; blen:BYTE; ans:PBYTE; pabitlen:PBYTE; anspar:PBYTE):DWORD; stdcall;
    _s_mfDataRead       = function(pccr:PCCR; timeout:BYTE; mfdm:BYTE; acm:BYTE; ak:BYTE;
  start:BYTE; len:BYTE; pdata:PBYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE):DWORD; stdcall;
    _s_mfDataWrite      = function(pccr:PCCR; timeout:BYTE; mfdm:BYTE; acm:BYTE; ak:BYTE;
  start:BYTE; len:BYTE; pdata:PBYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE):DWORD; stdcall;

var
// * global fn from library *
//  link_isopen(pccr:PCCR), test device handle
    link_isopen:        _s_link_isopen;
// link_close(pccr:PCCR), close device handle
    link_close:         _s_link_close;
// link_hidopen(idx:DWORD; pusn:PDWORD; pccr:PCCR):DWORD, open USB device; return LERR_SUCCESS, LERR_HARDWARE
    link_hidopen:       _s_link_hidopen;
// link_comopen(port:DWORD; cbr:DWORD; pccr:PCCR):DWORD, open COM port; return LERR_SUCCESS, LERR_HARDWARE
    link_comopen:       _s_link_comopen;
// link_flush(pccr:PCCR):DWORD, queues flush; return LERR_SUCCESS, LERR_HARDWARE
    link_flush:         _s_link_flush;
// link_echobyte(pccr:PCCR; dt:BYTE):DWORD, byte transmitting then echo receiving;
// return LERR_SUCCESS, LERR_HARDWARE, LERR_TIMEOUT_RX, LERR_TIMEOUT_TX, LERR_NOTSYNC
    link_echobyte:      _s_link_echobyte;
// link_rqwait(pccr:PCCR; buf:PBYTE; len:BYTE):DWORD, wait for request from reader;
// return LERR_SUCCESS, LERR_HARDWARE, LERR_NOTSYNC, LERR_TIMEOUT_RX
    link_rqwait:        _s_link_rqwait;
// link_rqack(pccr:PCCR; ack:BYTE):DWORD, send acknowledge to reader;
// return LERR_SUCCESS, LERR_HARDWARE, LERR_NOTSYNC, LERR_TIMEOUT_RX, LERR_TIMEOUT_TX
    link_rqack:         _s_link_rqack;
// link_packet(pccr:PCCR; cmd:BYTE; parm:BYTE; outbuf:PBYTE; outlen:WORD; inbuf:PBYTE; inlen:WORD):DWORD,
// main function for reader IO
    link_packet:        _s_link_packet;
// acq2type(acq:DWORD; atslen:BYTE; ats:PBYTE):BYTE, convert uidlen, atqa, sak to card type
    acq2type:           _s_acq2type;
// mfInit(pccr:PCCR; pmodel:PBYTE; phw:PBYTE; psn:PDWORD):DWORD, DIR-mode initialization
    mfInit:             _s_mfInit;
// mfBeep(pccr:PCCR; tone:BYTE):DWORD, beep, tone = 0..5
    mfBeep:             _s_mfBeep;
// mfBrkCarrier(pccr:PCCR; time_ms:WORD):DWORD, carrier break
    mfBrkCarrier:       _s_mfBrkCarrier;
// mfType(pccr:PCCR; timeout:BYTE; ptype:PBYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE; patslen:PBYTE; pats:PBYTE):DWORD,
// card type detection
    mfType:             _s_mfType;
// mfSelect(pccr:PCCR; timeout:BYTE; rq:BYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE):DWORD,
// select card
    mfSelect:           _s_mfSelect;
// mfSetAK(pccr:PCCR; ak:BYTE; key:PBYTE):DWORD, set authorization key
    mfSetAK:            _s_mfSetAK;
// mfAuth(pccr:PCCR; timeout:BYTE; acm:BYTE; ak:BYTE; block:BYTE; puidlen:PBYTE; puid:PBYTE; patqa:PWORD; psak:PBYTE):DWORD,
// data block authorization only then card halted
    mfAuth:             _s_mfAuth;
// mfLong(pccr:PCCR; txbuf:PBYTE; txlen:BYTE; ans:PBYTE; pabitlen:PBYTE):DWORD,
// send long command and receive answer
    mfLong:             _s_mfLong;
// mfCalcCrcA(buf:PBYTE; len:DWORD):WORD, calculate CRC-A and append buf
    mfCalcCrcA:         _s_mfCalcCrcA;
// mfAppendCrcA(buf:PBYTE; len:DWORD), calculate CRC-A and append buf
    mfAppendCrcA:       _s_mfAppendCrcA;
// mfIO(pccr:PCCR; buf:PBYTE; len:BYTE; par:PBYTE; ans:PBYTE; pabitlen:PBYTE; anspar:PBYTE):DWORD,
// basis function for card IO
    mfIO:               _s_mfIO;
// mfBitIO(pccr:PCCR; buf:PBYTE; blen:BYTE; ans:PBYTE; pabitlen:PBYTE; anspar:PBYTE):DWORD,
// basis function for card IO (bit based)
    mfBitIO:            _s_mfBitIO;

// *** read data block/page ***
// pccr    - pointer to the handle struct (see link_hidopen or link_comopen)
// timeout - operation timeout 0..7, time_in_ms=(timeout*4+3)*25; min=75mS, max=775mS
// mfdm    - data mode: MFDM_BLOCK | MFDM_PAGE (block/page mode), MFDM_NMC | MFDM_RWC (normal/RW card)
// acm     - authority command (PICC_AUTHENT1A | PICC_AUTHENT1B): no authority mode when acm=0
//  ak     - authority key 0..84, RAM-key when ak=0
// start   - start block or page for read 0..255
// len     - data length in blocks or pages
// pdata   - a pointer to the data buffer, may be NULL for 'dummy' read operation
// puidlen - may be NULL; before call: length of UID in puid for force operation with UID = puid ONLY!
//           puidlen must be NULL or *puidlen must be 0 if need 'any UID' mode
//           after call: the length of the UID card used
// puid    - may be NULL; before call: UID for force operation with card UID = puid ONLY!
//           after call: UID of last used card
// patqa   - may be NULL; after call: ATQA of last used card
// psak    - may be NULL; after call: SAK of last used card
// return  - error code
    mfDataRead:         _s_mfDataRead;

// *** write data block/page ***
// pccr    - pointer to the handle struct (see link_hidopen or link_comopen)
// timeout - operation timeout 0..7, time_in_ms=(timeout*4+3)*25; min=75mS, max=775mS
// mfdm    - data mode: MFDM_BLOCK | MFDM_PAGE (block/page mode), MFDM_NMC | MFDM_RWC (normal/RW card)
// acm     - authority command (PICC_AUTHENT1A | PICC_AUTHENT1B): no authority mode when acm=0
//  ak     - authority key 0..84, RAM-key when ak=0
// start   - start block or page for read 0..255
// len     - data length in blocks or pages
// pdata   - a pointer to the data buffer, can't be NULL
// puidlen - may be NULL; before call: length of UID in puid for force operation with UID = puid ONLY!
//           puidlen must be NULL or *puidlen must be 0 if need 'any UID' mode
//           after call: the length of the UID card used
// puid    - may be NULL; before call: UID for force operation with card UID = puid ONLY!
//           after call: UID of last used card
// patqa   - may be NULL; after call: ATQA of last used card
// psak    - may be NULL; after call: SAK of last used card
// return  - error code
    mfDataWrite:        _s_mfDataWrite;

// ******************************************************
// *** end of GLOBAL constant & vars for rd0xAB32.dll ***
// ******************************************************

label                   main_exit;

// *** load CCR library ***
function loadlib_rd0xab(var hLibRD0xAB:HModule):DWORD;
begin
  Result := LERR_SUCCESS;
  // load library
  hLibRD0xAB := LoadLibrary(RD0XAB_NM);
  if (hLibRD0xAB <> 0) then
  begin
    // load pointer to fns
    @link_isopen := GetProcAddress(hLibRD0xAB, PChar(LINK_ISOPEN_NM));
    @link_close := GetProcAddress(hLibRD0xAB, PChar(LINK_CLOSE_NM));
    @link_hidopen := GetProcAddress(hLibRD0xAB, PChar(LINK_HIDOPEN_NM));
    @link_comopen := GetProcAddress(hLibRD0xAB, PChar(LINK_COMOPEN_NM));
    @link_flush := GetProcAddress(hLibRD0xAB, PChar(LINK_FLUSH_NM));
    @link_echobyte := GetProcAddress(hLibRD0xAB, PChar(LINK_ECHOBYTE_NM));
    @link_rqwait := GetProcAddress(hLibRD0xAB, PChar(LINK_RQWAIT_NM));
    @link_rqack := GetProcAddress(hLibRD0xAB, PChar(LINK_RQACK_NM));
    @link_packet := GetProcAddress(hLibRD0xAB, PChar(LINK_PACKET_NM));
    @acq2type := GetProcAddress(hLibRD0xAB, PChar(ACQ2TYPE_NM));
    @mfInit := GetProcAddress(hLibRD0xAB, PChar(MFINIT_NM));
    @mfBeep := GetProcAddress(hLibRD0xAB, PChar(MFBEEP_NM));
    @mfBrkCarrier := GetProcAddress(hLibRD0xAB, PChar(MFBRKCARRIER_NM));
    @mfType := GetProcAddress(hLibRD0xAB, PChar(MFTYPE_NM));
    @mfSelect := GetProcAddress(hLibRD0xAB, PChar(MFSELECT_NM));
    @mfSetAK := GetProcAddress(hLibRD0xAB, PChar(MFSETAK_NM));
    @mfAuth := GetProcAddress(hLibRD0xAB, PChar(MFAUTH_NM));
    @mfLong := GetProcAddress(hLibRD0xAB, PChar(MFLONG_NM));
    @mfCalcCrcA := GetProcAddress(hLibRD0xAB, PChar(MFCALCCRCA_NM));
    @mfAppendCrcA := GetProcAddress(hLibRD0xAB, PChar(MFAPPENDCRCA_NM));
    @mfIO := GetProcAddress(hLibRD0xAB, PChar(MFIO_NM));
    @mfBitIO := GetProcAddress(hLibRD0xAB, PChar(MFBITIO_NM));
    @mfDataRead := GetProcAddress(hLibRD0xAB, PChar(MFDATAREAD_NM));
    @mfDataWrite := GetProcAddress(hLibRD0xAB, PChar(MFDATAWRITE_NM));
    // check pointers
    if  not assigned (link_isopen) or
        not assigned (link_close) or
        not assigned (link_hidopen) or
        not assigned (link_comopen) or
        not assigned (link_flush) or
        not assigned (link_echobyte) or
        not assigned (link_rqwait) or
        not assigned (link_rqack) or
        not assigned (link_packet) or
        not assigned (acq2type) or
        not assigned (mfInit) or
        not assigned (mfBeep) or
        not assigned (mfBrkCarrier) or
        not assigned (mfType) or
        not assigned (mfSelect) or
        not assigned (mfSetAK) or
        not assigned (mfAuth) or
        not assigned (mfLong) or
        not assigned (mfCalcCrcA) or
        not assigned (mfAppendCrcA) or
        not assigned (mfIO) or
        not assigned (mfBitIO) or
        not assigned (mfDataRead) or
        not assigned (mfDataWrite) then Result := LERR_FAIL;
  end;
end;

const
  // default keys
  defKeysA:     packed array[0 .. 2, 0 .. 5] of BYTE =(
  ($FF, $FF, $FF, $FF, $FF, $FF),
  ($00, $00, $00, $00, $00, $00),
  ($A0, $A1, $A2, $A3, $A4, $A5) );
  defKeysB:     packed array[0 .. 2, 0 .. 5] of BYTE =(
  ($FF, $FF, $FF, $FF, $FF, $FF),
  ($00, $00, $00, $00, $00, $00),
  ($B0, $B1, $B2, $B3, $B4, $B5) );

var
  hLibRD0xAB:   HModule = 0;
  ccr:          TCCR = ( hL: $FFFFFFFF );
  err:          DWORD = LERR_SUCCESS;
  sc_model:     BYTE = 0;
  sc_hw:        BYTE = 0;
  sc_sn:        DWORD = 0;

  sc_tp:        BYTE = SCD_UNKNOWN;
  sc_uidlen:    BYTE;
  sc_uid:       packed array[0 .. SZ_UID - 1] of BYTE;
  sc_atqa:      WORD;
  sc_sak:       BYTE;
  sc_atslen:    BYTE;
  sc_ats:       packed array[0 .. SZ_ATS - 1] of BYTE;

  pg, pglen:    BYTE;

  blk, blklen:  BYTE;
  blk_acm:      BYTE;
  blkf:         BOOL;

  buf:          packed array[0 .. SZ_MFBLK * 4 - 1] of BYTE;

  strLocal:     String;
  cnt:          DWORD;

// main entry
begin
  Writeln('rd0xAB read card demo(use rd0xAB.dll), API V1.6');
  Writeln('--');
  // connect to library rd0xAB.dll
  err := loadlib_rd0xab(hLibRD0xAB);
  if err <> LERR_SUCCESS then goto main_exit;

  // connecting to reader
  Write('Connecting... ');
  // open first USB device
  err := link_hidopen(0, nil, @ccr);
  if  err <> LERR_SUCCESS then goto main_exit;
  // initialize DIR mode
  err := mfInit(@ccr, @sc_model, @sc_hw, @sc_sn);
  if  err <> LERR_SUCCESS then goto main_exit;
  strLocal := Format('Ok. RD-%.02uAB, SN:%.08X, HW:%u.%u.', [sc_model, sc_sn, sc_hw shr 4, sc_hw and $0F]);
  Writeln(strLocal);
  Writeln('--');

  // *** determine card type ***
  write('Card type... ');
  sc_uidlen := MFDM_RWC;
  err := mfType(@ccr, 3, @sc_tp, @sc_uidlen, @sc_uid, @sc_atqa, @sc_sak, @sc_atslen, @sc_ats);
  if err <> LERR_SUCCESS then goto main_exit;
  // print card type
  if sc_tp = SCD_UL then Write('UL')                // ultralight
  else if sc_tp = SCD_CL1K then Write('CL1K')       // Classic 1K
  else if sc_tp = SCD_CL4K then Write('CL4K')       // Classic 4K
  else if sc_tp = SCD_CL1K1 then Write('CL1K1')     // Classic 1K
  else if sc_tp = SCD_CL4K1 then Write('CL4K1')     // Classic 4K
  else if sc_tp = SCD_PL2S1 then Write('PL2S1')     // Plus 2K, Sec.Lev.1
  else if sc_tp = SCD_PL4S1 then Write('PL4S1')     // Plus 4K, Sec.Lev.1
  else if sc_tp = SCD_PL2S2 then Write('PL2S2')     // Plus 2K, Sec.Lev.2
  else if sc_tp = SCD_PL4S2 then Write('PL4S2')     // Plus 4K, Sec.Lev.2
  else if sc_tp = SCD_PL2S3 then Write('PL2S3')     // Plus 2K, Sec.Lev.3
  else if sc_tp = SCD_PL4S3 then Write('PL4S3')     // Plus 4K, Sec.Lev.3
  else if sc_tp = SCD_MINI then Write('MINI')       // Classic Mini
  else if sc_tp = SCD_DF1 then Write('DF1')         // DesFire, EV1
  else if sc_tp = SCD_DF2 then Write('DF2')         // DesFire, EV2
  else if sc_tp = SCD_SMMX then Write('SMMX')       // Smart MX
  else if sc_tp = SCD_SM1K then Write('SM1K')       // Smart MX 1K
  else if sc_tp = SCD_SM2K then Write('SM2K')       // Smart MX 2K
  else if sc_tp = SCD_SM4K then Write('SM4K')       // Smart MX 4K
  else Write('???');                                // ?
  if (sc_uidlen and MFDM_RWC) <> 0 then Write('-RW');   // set RW flag
  Writeln('.');
  // print UID
  Write(' UID=');
  if (sc_uidlen and $0F) <> 0 then
    for cnt := 0 to (sc_uidlen and $0F) - 1 do
    begin
      strLocal := Format('%.02X',[sc_uid[cnt]]);
      Write(strLocal);
    end;
  // print ATQA, SAK
  strLocal := Format(', ATQA=%.04X, SAK=%.02X.', [sc_atqa, sc_sak]);
  Writeln(strLocal);
  // print ATS
  Write(' ATS=');
  if sc_atslen = 0 then Write('none')
  else if sc_atslen = 1 then
  begin
    strLocal := Format('none (ack=%.02X)', [sc_ats[0]]);
    Write(strLocal);
  end
  else
  begin
    for cnt := 0 to sc_atslen - 1 do
    begin
      strLocal := Format('%.02X',[sc_ats[cnt]]);
      Write(strLocal);
    end;
  end;
  Writeln('.');
  Writeln('--');

  // *** read card ***
  if sc_tp = SCD_UL then
  begin
    pg := 0;
    pglen := 16;
    // *** read pages from UL card ***
    strLocal := Format('Reading pages from %u to %u... ',[pg, pg + pglen - 1]);
    Write(strLocal);
    err := mfDataRead(@ccr, 7, MFDM_NMC or MFDM_PAGE,
      0, 0, pg, pglen, @buf, nil, nil, nil, nil);
    if err <> LERR_SUCCESS then goto main_exit;
    Write('Ok.');
    // print data
    for cnt := 0 to pglen * SZ_MFPAGE - 1 do
    begin
      if (cnt mod 16) = 0 then
      begin
        Writeln;
        write(' ');
      end;
      strLocal := Format('%.02X', [buf[cnt]]);
      Write(strLocal);
    end;
    Writeln;
    Writeln('--');
  end
  else if (sc_tp = SCD_CL1K) or
          (sc_tp = SCD_CL4K) or
          (sc_tp = SCD_CL1K1) or
          (sc_tp = SCD_CL4K1) or
          (sc_tp = SCD_PL2S1) or
          (sc_tp = SCD_PL4S1) or
          (sc_tp = SCD_SM1K) or
          (sc_tp = SCD_SM4K) then
  begin
    blk := 0;
    blklen := 4;
    blkf := FALSE;
    // *** read blocks from CL/PL/SM card ***
    strLocal := Format('Reading blocks from %.02X to %.02X...',[blk, blk + blklen - 1]);
    Writeln(strLocal);
    // key A auth
    for cnt := 0 to 2 do
    begin
      strLocal := Format('Read with key A[%.02X%.02X%.02X%.02X%.02X%.02X]... ',
        [defKeysA[cnt][0], defKeysA[cnt][1], defKeysA[cnt][2],
         defKeysA[cnt][3], defKeysA[cnt][4], defKeysA[cnt][5]]);
      Write(strLocal);
      // * set new authorization key in RAM(key 0) *
      err := mfSetAK(@ccr, 0, @defKeysA[cnt][0]);
      if err <> LERR_SUCCESS then goto main_exit;
      // * read data with key 0 *
      blk_acm := PICC_AUTHENT1A;  // auth A cmd
      sc_uidlen := 0;             // set 0 for read any UID
      err := mfDataRead(@ccr, 7, MFDM_NMC or MFDM_BLOCK,
        blk_acm, 0, blk, blklen, @buf, @sc_uidlen, nil, nil, nil);
      if (err <> LERR_SUCCESS) and (err < LERR_NOCARD) then goto main_exit;
      if err <> LERR_SUCCESS then Writeln('Fail.')
      else
      begin
        blkf := TRUE;
        Writeln('Ok.');
        Break;
      end;
    end;
    // key B auth
    if blkf = FALSE then
    begin
      for cnt := 0 to 2 do
      begin
        strLocal := Format('Read with key B[%.02X%.02X%.02X%.02X%.02X%.02X]... ',
          [defKeysB[cnt][0], defKeysB[cnt][1], defKeysB[cnt][2],
          defKeysB[cnt][3], defKeysB[cnt][4], defKeysB[cnt][5]]);
        Write(strLocal);
        // * set new authorization key in RAM(key 0) *
        err := mfSetAK(@ccr, 0, @defKeysB[cnt][0]);
        if err <> LERR_SUCCESS then goto main_exit;
        // * read data with key 0 *
        blk_acm := PICC_AUTHENT1B;  // auth B cmd
        sc_uidlen := 0;             // set 0 for read any UID
        err := mfDataRead(@ccr, 7, MFDM_NMC or MFDM_BLOCK,
          blk_acm, 0, blk, blklen, @buf, @sc_uidlen, nil, nil, nil);
        if (err <> LERR_SUCCESS) and (err < LERR_NOCARD) then goto main_exit;
        if err <> LERR_SUCCESS then Writeln('Fail.')
        else
        begin
          blkf := TRUE;
          Writeln('Ok.');
          Break;
        end;
      end;
    end;
    if blkf <> FALSE then
    begin
      // print data
      for cnt := 0 to blklen * SZ_MFBLK - 1 do
      begin
        if (cnt mod 16) = 0 then
        begin
          Writeln;
          write(' ');
        end;
        strLocal := Format('%.02X', [buf[cnt]]);
        Write(strLocal);
      end;
      Writeln;
      Writeln('--');
    end
    else
    begin
      Writeln('Can''t read data from card!');
      Writeln('--');
     err := LERR_SUCCESS;
    end;
  end;
  Writeln('Operation complete.');

main_exit:

  // verbose error code
  if err = LERR_FAIL then Writeln('Command execution error!')
  else if err = LERR_HARDWARE then Writeln('Hardware error!')
  else if err = LERR_TIMEOUT_TX then Writeln('Communication error: TX timeout!')
  else if err = LERR_TIMEOUT_RX then Writeln('Communication error: RX timeout!')
  else if err = LERR_NOTSYNC then Writeln('Communication error: synchronization lost!')
  else if err = LERR_NOREADER then Writeln('Can''t find reader!')
  else if err = LERR_NOCARD then Writeln('Can''t find card!')
  else if err = LERR_INVALIDCARD then Writeln('Invalid card type!')
  else if err = LERR_INVALIDKEY then Writeln('Invalid auth key or card type!')
  else if err = LERR_IOERR then Writeln('I/O error!')
  else if err = LERR_DATAREAD then Writeln('Can''t read data from card!')
  else if err = LERR_DATAWRITE then Writeln('Can''t write data to card!')
  else if err = LERR_BUSY then Writeln('Device busy error!')
  else if err = LERR_USER then Writeln('Unknown error!');

  // *** beep! ***
  if (ccr.hL <> INVALID_HANDLE_VALUE) and
     (err <> LERR_HARDWARE) and
     (err <> LERR_NOREADER) then
  if err <> LERR_SUCCESS then mfBeep(@ccr, 3)     // if fail
  else  mfBeep(@ccr, 2);                          // if success

  // close USB handle
  link_close(@ccr);

  // disconnect from library rd0xAB.dll
  if (hLibRD0xAB <> 0) then FreeLibrary(hLibRD0xAB);

  Sleep(5000);
end.

